'''Pybat is a module designed to make it easy to use python in place of batch
files, and to share environment variables between python scripts and batch
files.

Author: Albert Sweigart

Note: To share environment variables, pybat.py must have file creation permissions
for the current directory.

Note: pybat.py and pybat.bat MUST be in the same directory.

* Copyright (c) 2008, Albert Sweigart
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the PyBat nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Albert Sweigart ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Albert Sweigart BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

# Build 1

import os, sys, copy, random

def main():
    '''pybat.py can be called in the following modes (sys.argv[1] will equal the mode name):
    generatenonce - returns a file nonce as the exit code
    outputenvvar - writes out the environment variables to envvar(nonce).bat file (no exit code)
    callpy - calls a python script (parses command line and resets sys.argv values), returns script's exit code as pybat's exit code
    '''

    if len(sys.argv) >= 2:
        if sys.argv[1] == 'generatenonce':
            sys.exit(generateFileNonce())

        elif sys.argv[1].startswith('chopparams'):
            sys.exit(generateParamTemp(sys.argv[1][10:]))

        elif sys.argv[1] == 'outputenvvar':
            if len(sys.argv) >= 3:
                outputEnvVars(sys.argv[2])
                sys.exit()

        elif sys.argv[1] == 'callpy':
            if len(sys.argv) >= 4:
                exitcode = callpy(' '.join(sys.argv[3:]))
                outputEnvVars(sys.argv[2])
                sys.exit(exitcode)

    usage()
    sys.exit(1) # generic error code.

def usage():
    print 'POSSIBLE ERROR - pybat.py not called with correct command.'
    print 'Usage: pybat.py [command] [option(s)]'
    print
    print 'Commands [with options]:'
    print '\tgeneratenonce - returns a file nonce as the exit code'
    print '\toutputenvvar filenonce - writes out the environment variables to envvar(nonce).bat file (no exit code)'
    print "\tcallpy filenonce python_script_filename [args] - calls a python script returns script's exit code as pybat's exit code"
    print

def parseCmdline(cmdline):
    '''Takes a string command line, expands any environment variables, and then parses the file run and it's arguments.
    Returns a 2-tuple, whose first item is the string of the file to be run, and second item is a list of strings of arguments.'''

    # First, expand any environment variables
    cmdline = winexpandvars(cmdline)

    # for cmd.exe, single-quotes are not the same as double-quotes, and do not expand environment variables

    # Grab the first part as the filename
    if cmdline.strip()[0] == '"':
        # first part is not surrounded by quotes
        endQuote = cmdline[1:].find('"')
        if endQuote == -1:
            # there is a start quote but no ending quote. Technically this is legal, and is equivalent to having the entire line in quotes (and no arguments)
            cmdline += '"'
            args = []
        else:
            args = cmdline[endQuote+2:].split() # endQuote+2 is just after the endQuote in the full path
            cmdline = cmdline[0:endQuote+2]
    else:
        args = cmdline.split()[1:]
        cmdline = '"' + cmdline.split()[0] + '"'

    return (cmdline, args)


def callpy(cmdline):
    '''Calls a python script with arguments, and returns the script's exit code.
    This function is run when a batch file calls pybat.bat which then calls pybat.py, passing pyfilename.

    pyfilename - string of the path to the python script that is to be run
    args - list of strings of arguments to the python script

    Returns exit code of the called python script'''

    # TODO need to be able to pass env vars as part of the python script filename.
    fileToRun, args = parseCmdline(cmdline)
    # since we are passing fileToRun to execfile(), remove the double quotes surrounding the name
    fileToRun = fileToRun.replace('"', '')

    retval = 0
    storedArgs = copy.copy(sys.argv) # store current sys.argv values
    sys.argv[0] = os.path.abspath(fileToRun)
    sys.argv[1:] = args
    try:
        execfile(fileToRun)
    except SystemExit, exitcode:
        retval = str(exitcode)
    except Exception, excobj:
        raise

    sys.argv = storedArgs # restore sys.argv values

    return int(retval)


def callbat(cmdline):
    '''Calls a pybat.bat to call batch file with arguments, and returns the batch file's exit code.
    This function is called by python scripts which have loaded the pybat module.
    This function is run when a python script runs pybat.py, and then this function runs pybat.bat to run the batch file.

    batfilename - string of the batch file's filename

    Returns exit code of the called batch file'''

    # TODO need to be able to pass env vars as part of the batch filename.
    # TODO check that .cmd files can be run as well.
    fileToRun, args = parseCmdline(cmdline)

    filenonce = generateFileNonce()
    os.system('pybat.bat callbat %s %s %s' % (filenonce, fileToRun, ' '.join(args)))
    # os.system() only returns exit codes of 0 from batch file calls even with cmd.exe, despite what the documentation says for Windows.
    readEnvVarFile(filenonce) # loads the environment variables (including any changes the called batch file made)
    exitcode = int(os.environ['pybatexitcode2'])
    os.unlink('envvar%s.bat' % (filenonce))
    return exitcode


def winexpandvars(s):
    """Expand environment variables of form %var%.

    Unknown variables are left unchanged.
    """

    if '%' not in s:
        return s

    ptr = 0
    startPercent = 0
    inPercentState = False
    while ptr < len(s):
        if s[ptr] == '%' and not inPercentState:
            inPercentState = True
            startPercent = ptr
        elif s[ptr] == '%' and inPercentState:
            if s[startPercent+1:ptr] in os.environ:
                newStartPercent = startPercent + len(os.environ[s[startPercent+1:ptr]]) - 1
                s = s[:startPercent] + os.environ[s[startPercent+1:ptr]] + s[ptr+1:]
                ptr = newStartPercent
                inPercentState = False
            else:
                startPercent = ptr
        ptr += 1
    return s


def outputEnvVars(filenonce):
    '''Creates a file named envvar(filenonce).bat, writes out the current environment variables to it.

    Returns nothing.'''
    fp = open('envvar%s.bat' % (filenonce), 'w')
    for i in os.environ:
        fp.write('set %s=%s\n' % (i, os.environ[i]))
    fp.close()


def generateFileNonce():
    '''Figures out a unique filename in the current directory to create a new
    file with a filename in the format of "envvarNONCE.bat".

    Returns a string of the filename's nonce.'''
    while True:
        nonce = random.randint(2,2147483647)
        if not os.path.exists('envvar%s.bat' % (nonce)):
            # NOTE: Possible concurrency problem here if another instance of pybat runs, since the above and below lines are not atomic.
            fp = open('envvar%s.bat' % (nonce), 'w')
            fp.close()
            return nonce


def generateParamTemp(numToChop):
    '''Creates a file that has parameters %3 and on in a file of the
    format "pybatparamtempEXITCODE" (no file extension)

    Returns int of the exit code.'''

    numToChop = int(numToChop)

    cnt = 2 # We use an exit code of 1 as an error notification, so let's always use 2 and up for the nonce.
    while True:
        if not os.path.exists('pybatparamtemp%s' % (cnt)):
            fp = open('pybatparamtemp%s' % (cnt), 'w')
            # ensure that sys.argv[numToChop] is enclosed in double quotes.
            sys.argv[numToChop] = sys.argv[numToChop].strip('\'\"')
            sys.argv[numToChop] = '"' + sys.argv[numToChop] + '"'
            fp.write(' '.join(sys.argv[numToChop:]))
            fp.close()
            return cnt
        cnt += 1


def readEnvVarFile(filenonce):
    '''Reads through the envvar(nonce).bat file and loads the environment variables in there into os.environ.

    filenonce - integer or string of the filenonce, which tells readEnvVarFile() which file to read.

    Returns nothing.'''
    try:
        fp = open('envvar%s.bat' % (filenonce))
    except:
        print 'Unable to open envvar%s.bat to read in environment variables.' % (filenonce)
        raise
    content = fp.readlines()
    fp.close()

    for line in content:
        ind = line.find('=')
        if '$' in line:
            pass
        if ind != -1:
            os.environ[line[4:ind]] = line[ind+1:].rstrip('\n')


# ======================================================================================================
# These functions are just helpful utilities functions, but they aren't called by pybat.py or pybat.bat:

def env2py(vars = None):
    '''Creates python variables in the global scope from environment variables.
    vars can be None, in which case all python variables are copied into python variables
    vars can be a string of one variable name, which converts that variable into an python variable
    vars can be a list of strings of variable names, which converts those variables into python variables

    Returns nothing.'''
    if vars is None:
        vars = os.environ
    elif type(vars) == type(''):
        vars = [x for x in [vars] if x in os.environ]
    elif type(vars) == type([]):
        vars = [x for x in vars if x in os.environ]

    for i in vars:
        eval('%s=%s' % (i, os.environ[i]))


def py2env(vars = None):
    '''Sets environment variables.
    vars can be None, in which case all python variables are copied into environment variables
    vars can be a string of one variable name, which converts that variable into an environment variable
    vars can be a list of strings of variable names, which converts those variables into environment variables

    Returns nothing.'''
    if vars is None:
        vars = []
        globcopy = copy.copy(globals())
        for i in globcopy:
            if i not in ['__builtins__', '__name__', '__file__', '__doc__', 'vars', 'globcopy'] and not str(globals()[i].startswith('<module')):
                vars.append(i)
    elif type(vars) == type(''):
        vars = [x for x in [vars] if x in globals()]
    elif type(vars) == type([]):
        vars = [x for x in vars if x in globals()]

    for i in vars:
        eval('os.environ[i] = %s' % (i)) # TODO: possible security risk if environment variables can be formatted as python code


def filesin(filepath):
    '''Return a list of the absolute filepaths of files in filepath (does not
    return '.' and '..')'''
    if not os.path.exists(os.path.abspath(filepath)):
        return []

    return sorted(filter(os.path.isfile, os.listdir(filepath)))


def dirsin(filepath):
    '''Return a list of the absolute filepaths of directories in filepath'''
    if not os.path.exists(os.path.abspath(filepath)):
        return []

    return sorted(filter(os.path.isdir, os.listdir(filepath)))


def subfilesin(filepath):
    '''Return a list of the absolute filepaths of files in filepath and its
    subdirectories (does not return '.' and '..')'''
    if not os.path.exists(os.path.abspath(filepath)):
        return []

    for dirpath, dirnames, filenames in os.walk('.'):
        temp = []
        for i in filenames:
            temp.append(os.path.abspath(dirpath + i))
        retval.extend(sorted(temp))
    return retval


def subdirsin(filepath):
    '''Return a list of the absolute filepaths of directories in filepath and
    its subdirectories'''
    if not os.path.exists(os.path.abspath(filepath)):
        return []

    for dirpath, dirnames, filenames in os.walk('.'):
        temp = []
        for i in dirnames:
            temp.append(os.path.abspath(dirpath + i))
        retval.extend(sorted(temp))
    return retval


if __name__ == '__main__':
    main()